# built-ins imports
import shutil
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor, as_completed

# custom imports
from pint import Quantity
from rich import print
from rich.text import Text
from rich.console import Console, Group
from rich.progress import Progress, BarColumn, TextColumn
from rich.panel import Panel, Padding
from rich.live import Live
from rich.prompt import Prompt

from pandas import DataFrame, Series


class FileTransferManager:

    results: DataFrame
    direction: str
    sessions: DataFrame

    def __init__(self, sessions, results=None, direction=None):

        if isinstance(sessions, Series):
            sessions = sessions.to_frame().transpose()

        self.sessions = sessions
        self.results = results  # type: ignore
        self.direction = direction  # type: ignore

    @staticmethod
    def from_transfer(transfer):
        return FileTransferManager(transfer.sessions, transfer.results, transfer.direction)

    def _assert_file_checked(self, function_name):
        if self.results is None or self.direction is None:
            raise ValueError(
                f"Cannot show {function_name} if no result has been "
                "obtained through a fetch or a push_request command"
            )

    def resolve(self):

        results = self.results.copy()
        conflicts = results[results["decision"] == "conflict"]
        accepted_decisions = ["transfer", "overwrite", "ignore", "conflict"]

        def conflict_pannel(conflict=None):

            if conflict is None:
                return Panel("", title="❗ Handling Conflict :", border_style="dark_blue")

            line = Text.assemble(
                ("📄"),
                (f"{conflict.source_filepath}\n", "steel_blue1 reverse"),
                ("⏬\n"),
                (f"{conflict.destination_filepath}\n", "steel_blue1 reverse"),
                ("Informations : 🚩:", "dark_blue underline"),
                (f"{conflict.warnings}\n\n", "dark_blue"),
                (f"Type in your decision. (One of {accepted_decisions})"),
            )

            return Panel(line, title="❗ Handling Conflict :", border_style="dark_blue")

        if len(conflicts):
            console = Console()
            with Live(conflict_pannel(), console=console, refresh_per_second=5) as live:
                for index, conflited_file in results[results["decision"] == "conflict"].iterrows():
                    live.update(conflict_pannel(conflited_file))

                    decision = Prompt.ask(choices=accepted_decisions)
                    if decision not in accepted_decisions:
                        raise ValueError(f"Must be one of {accepted_decisions}, got {decision}")
                    else:
                        results.loc[index, "decision"] = decision

                conflicts = results[results["decision"] == "conflict"]
                if len(conflicts):
                    print(
                        Text(
                            "🚫 There seem to still exists conflicts to resolve. "
                            "Please run resolve again to fix the last ones 🚫",
                            style="bright_red bold",
                        )
                    )
                else:
                    print(Text("All conflicts solved ! 🎉", style="spring_green3 bold"))

        else:
            print(Text("No conflicts to resolve. ✅", style="spring_green3 bold"))

        return FileTransferManager(self.sessions, results=results, direction=self.direction)

    def status(self, show_message=True, return_messages=False):
        self._assert_file_checked("status")

        action = self.direction

        is_status_ok = True
        messages = []

        n_ignored_files = len(self.results[self.results["decision"] == "ignore"])
        n_transfered_files = len(self.results[self.results["decision"] == "transfer"])
        n_overritten_files = len(self.results[self.results["decision"] == "overwrite"])
        n_conflicts = len(self.results[self.results["decision"] == "conflict"])

        messages.append(Text("❗ Conflicting files:", style="dark_blue underline bold"))
        if n_conflicts:
            is_status_ok = False
            messages.append(
                Text(
                    f"\t🚫 {n_conflicts} conflicting files were found.\n"
                    "\tPlease sort out the conflits with the `.resolve()` command.\n",
                    style="bold bright_red",
                )
            )
        else:
            messages.append(Text("\tNo conflicts were found ! 🎉\n", style="bold spring_green3"))

        metrics = self.transfer_metrics()
        messages.append(Text("📈 Transfer metrics:", style="dark_blue underline bold"))

        messages.append(Text(f"\t• {n_transfered_files} files will be newly transfered.", style="spring_green3"))
        messages.append(
            Text(
                f"\t• {n_overritten_files} files will overwrite the equivalent file in destination.",
                style="dark_orange",
            )
        )
        messages.append(
            Text(
                f"\t• {n_ignored_files} files will be ignored " "(not transfered from source to destination).\n",
                style="dark_orange",
            )
        )

        transfer_destinations = metrics.groupby("destination")

        messages.append(
            Text(
                f"💽 {action.capitalize()}ing to {len(transfer_destinations)} locations:",
                style="dark_blue underline bold",
            )
        )

        for destination, metric in transfer_destinations:

            destination = str(destination)
            sources = ", ".join(metric.source.astype(str))  # type: ignore
            transfer_space = metric.transfer_space.sum()
            free_space = metric.free_space.iloc[0]

            enough_space = free_space > transfer_space
            color = "spring_green3" if enough_space else "bright_red"
            summary = "✅ Enough free space ✅" if enough_space else "❌ Not enough space ! ❌"

            transfer_space_text = f"{transfer_space.to_compact():.2f~P} "
            free_space_text = f"{free_space.to_compact():.2f~P}"

            min_length = 11

            sources = sources.rjust(min_length)
            fill_size_source = "." * (len(sources) - (len(transfer_space_text)))
            fill_description_source = "." * (len(sources) - len("Source "))

            line = Text.assemble(
                (f"• {summary}\n", color),
                ("  "),
                ("Source", f"{color} underline"),
                (" "),
                (fill_description_source, "grey62"),
                (" 🆚 ", color),
                ("Destination\n", f"{color} underline"),
                (f"  {transfer_space_text}", f"{color} bold"),
                (fill_size_source, "grey62"),
                (" ⏩ ", color),
                (f"{free_space_text}\n  ", f"{color} bold"),
                (sources, f"{color} reverse"),
                (" ⏩ ", color),
                (f"{destination}\n", f"{color} reverse"),
            )

            messages.append(Padding(Panel(line, border_style=color), (0, 0, 0, 6)))

            if not enough_space:
                is_status_ok = False

        messages.append(Text("🌠 Conclusion:", style="dark_blue underline bold"))

        if is_status_ok:
            messages.append(Text(f"\t• ✅ Able to {action} the data.", style="spring_green3"))
            messages.append(
                Text(
                    f"\t• 📊 In total {metrics.transfer_space.sum().to_compact():.2f~P} of data will be " f"{action}ed",
                    style="spring_green3",
                )
            )
        else:
            messages.append(
                Text(
                    f"\t• ❌ Cannot {action} the data. Please sort out the issues mentionned above", style="bright_red"
                )
            )

        if show_message:
            group = Group(*messages)
            panel = Panel(
                group,
                title=f"📟 Pre-{action.capitalize()}ing Status Report",
                width=100,
                border_style="light_steel_blue3",
            )
            print(panel)

        if return_messages:
            return messages

        return is_status_ok

    def transfer_metrics(self):
        self._assert_file_checked("status")

        transfers_infos = []
        for (source, destination), transfers in self.results.groupby(["source_volume", "destination_volume"]):
            free_space = shutil.disk_usage(destination).free * Quantity("bytes")

            transfered_files = transfers[transfers["decision"] == "transfer"]
            overwritten_files = transfers[transfers["decision"] == "overwrite"]

            transfer_space = (
                overwritten_files["destination_filesize"] - overwritten_files["source_filesize"]
            ).sum() + transfered_files["source_filesize"].sum()

            session_nb = len(transfers.session.unique())
            files_nb = len(transfers)

            transfers_infos.append(
                dict(
                    source=source,
                    destination=destination,
                    free_space=free_space,
                    transfer_space=transfer_space,
                    session_nb=session_nb,
                    files_nb=files_nb,
                )
            )

        return DataFrame(transfers_infos)

    def _source_repositories(self):
        return self._repositories("source")

    def _destination_repositories(self):
        return self._repositories("destination")

    def _repositories(self, localisation="source"):
        self._assert_file_checked("sources_directories")
        if localisation not in ["source", "destination"]:
            raise ValueError("_repositories localisation must be 'source' or 'destination'")

        def replace_element(root_path, rel_path):
            return str(root_path).replace(str(rel_path), " ")

        if self.direction == "push":
            if localisation == "source":
                repo_key = "local_path"
            else:
                repo_key = "remote_path"
        else:
            if localisation == "source":
                repo_key = "local_path"
            else:
                repo_key = "local_path"

        root_paths = self.sessions.apply(lambda row: replace_element(row[repo_key], row["rel_path"]), axis=1)
        return root_paths.unique().tolist()

    def push(self):
        self._assert_file_checked("push")
        if self.direction != "push":
            raise ValueError("Cannot push files without doing first a prepush check")
        return self.transfer()

    def pull(self):
        self._assert_file_checked("pull")
        if self.direction != "pull":
            raise ValueError("Cannot pull files without doing first a fetch check")
        return self.transfer()

    def transfer(self):
        status = self.status(show_message=False, return_messages=False)
        if not status:
            raise ValueError(f"Cannot {self.direction} the data. Please check issues with `.status()`")

        transfers = self.results[(self.results["decision"] == "transfer") | (self.results["decision"] == "overwrite")]
        transfer_results = self.copy_files_with_progress(
            transfers.source_filepath, transfers.destination_filepath, transfers.decision
        )
        # TODO : write a message here if transfer errors, using the transfer_results.success column
        print(Text(f"🎉 Finished {self.direction}ing successfully ! 🎉", style="spring_green3"))
        return transfer_results

    @staticmethod
    def copy_files_with_progress(src_paths, dst_paths, decisions, max_workers=4):

        def copy_file(src, dst, decision, progress, task_id):

            src, dst = Path(src), Path(dst)

            message = f"Failed to copy {src} to {dst}."
            success = False

            if not src.exists():
                return src, dst, success, f"{message} Source doesn't exist"

            if dst.exists() and decision != "overwrite":
                return (
                    src,
                    dst,
                    success,
                    (
                        f"{message} "
                        "Decision was set to {decision} but the destination existed already. "
                        "Decision should have been set to overwrite"
                    ),
                )

            dst.parent.mkdir(exist_ok=True, parents=True)

            try:
                shutil.copy2(src, dst)
                message = f"Copied {src} to {dst}"
                success = True
            except Exception as e:
                message = f"{message}: Unexpected error : {e}"

            progress.advance(task_id)
            return src, dst, success, message

        total_files = len(src_paths)

        with Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            TextColumn("({task.completed}/{task.total})"),
        ) as progress:
            task_id = progress.add_task("Copying files...", total=total_files)

            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                copy_results = []
                futures = [
                    executor.submit(copy_file, src, dst, dec, progress, task_id)
                    for src, dst, dec in zip(src_paths, dst_paths, decisions)
                ]
                for future in as_completed(futures):
                    src, dst, success, message = future.result()
                    copy_results.append(
                        dict(source_filepath=src, destination_filepath=dst, success=success, message=message)
                    )
        return DataFrame(copy_results)

    def fetch(self, policies=None, show_status=True):
        results = self.check_files(source="remote_path", destination="local_path", policies=policies)
        new_file_namager = FileTransferManager(self.sessions, results, direction="pull")
        if show_status:
            new_file_namager.status()
        return new_file_namager

    def push_request(self, policies=None, show_status=True):
        results = self.check_files(source="local_path", destination="remote_path", policies=policies)
        new_file_namager = FileTransferManager(self.sessions, results, direction="push")
        if show_status:
            new_file_namager.status()
        return new_file_namager

    def check_files(self, source="remote_path", destination="local_path", policies=None):

        console = Console()

        default_policies = {
            "close_dates": "conflict",
            "destination_older": "overwrite",
            "no_file_exists": "transfer",
            "destination_younger": "ignore",
            "close_dates_threshold": 10,
        }

        if policies is not None:
            default_policies.update(policies)

        policies = default_policies

        def get_volume(full_path, rel_path):
            full_path, rel_path = str(Path(full_path)), str(Path(rel_path))
            root_path = Path(full_path.replace(rel_path, ""))
            return Path(root_path.drive)

        results = []

        with Progress(console=console) as progress:

            task = progress.add_task(f"Checking files for {len(self.sessions)} sessions", total=len(self.sessions))

            for _, session in self.sessions.iterrows():

                source_path = Path(str(session[source]))
                destination_path = Path(str(session[destination]))

                source_volume = get_volume(source_path, session["rel_path"])
                destination_volume = get_volume(destination_path, session["rel_path"])

                for root, dirs, files in source_path.walk():

                    for file in files:

                        source_filepath = root / file
                        source_filesize = source_filepath.stat().st_size * Quantity("bytes")
                        relative_filepath = source_filepath.relative_to(source_path)
                        destination_filepath = destination_path / relative_filepath

                        destination_exists = destination_filepath.exists()

                        source_stat = source_filepath.stat()

                        source_creation_date = source_stat.st_birthtime
                        source_modification_date = source_stat.st_mtime

                        source_date = max(source_creation_date, source_modification_date)

                        if destination_exists:

                            destination_stat = destination_filepath.stat()

                            destination_creation_date = destination_stat.st_birthtime
                            destination_modification_date = destination_stat.st_mtime

                            destination_filesize = destination_stat.st_size * Quantity("bytes")

                            destination_date = max(destination_creation_date, destination_modification_date)

                            if (difference := abs(source_date - destination_date)) < policies["close_dates_threshold"]:
                                decision = policies["close_dates"]
                                warnings = (
                                    f"The destination file and the source file last modification dates "
                                    f"are very close ({difference} sec difference). "
                                    "Double checking the file contents before transfer might be safer."
                                )

                            elif destination_date > source_date:
                                decision = policies["destination_younger"]
                                warnings = (
                                    "The destination file date is more recent than the source file one. "
                                    "Double checking the file contents before transfering is mandatory !"
                                )

                            else:
                                decision = policies["destination_older"]
                                warnings = "The destination file is older than source. "
                                "Transfering will most likely be okay."

                        else:
                            decision = policies["no_file_exists"]
                            warnings = "File is not existing on the destination, transfering is absolutely okay."
                            destination_date = None
                            destination_filesize = 0 * Quantity("bytes")

                        record = dict(
                            source_filepath=source_filepath,
                            destination_filepath=destination_filepath,
                            source_filesize=source_filesize,
                            destination_filesize=destination_filesize,
                            relative_filepath=relative_filepath,
                            destination_exists=destination_exists,
                            source_date=source_date,
                            destination_date=destination_date,
                            source_volume=source_volume,
                            destination_volume=destination_volume,
                            decision=decision,
                            warnings=warnings,
                            session=session.u_alias,
                        )

                        results.append(record)
                progress.advance(task)

        return DataFrame(results)



"""Decorators and small standalone functions for api module"""
import logging
import urllib.parse
from functools import wraps
from typing import Sequence, Union, Iterable, Optional, List, Tuple
from collections.abc import Mapping
import fnmatch
from datetime import datetime

import pandas as pd
from iblutil.io import parquet
import numpy as np
from packaging import version

import one.alf.exceptions as alferr
from one.alf.files import rel_path_parts, get_session_path, get_alf_path
from one.alf.spec import FILE_SPEC, regex as alf_regex
import one.alf.io as alfio
import re

logger = logging.getLogger(__name__)


def Listable(t):
    """Return a typing.Union if the input and sequence of input"""
    return Union[t, Sequence[t]]


def ses2records(ses: dict, int_id=False):
    """Extract session cache record and datasets cache from a remote session data record.

    Parameters
    ----------
    ses : dict
        Session dictionary from Alyx REST endpoint
    int_id : bool
        If True, the UUIDs are converted to two int64s

    Returns
    -------
    pd.Series
        Session record
    pd.DataFrame
        Datasets frame
    """
    # Extract session record
    eid = ses["url"][-36:]
    if int_id:
        eid = tuple(parquet.str2np(eid).flatten())
    session_keys = (
        "subject",
        "start_time",
        "lab",
        "number",
        "task_protocol",
        "projects",
    )
    session_data = {k: v for k, v in ses.items() if k in session_keys}
    session = pd.Series(data=session_data, name=eid).rename({"start_time": "date"})
    session["projects"] = ",".join(session.pop("projects"))
    session["date"] = datetime.fromisoformat(session["date"]).date()

    # Extract datasets table
    def _to_record(d):
        rec = dict(file_size=d["file_size"], hash=d["hash"], exists=True)
        if int_id:
            rec["id_0"], rec["id_1"] = parquet.str2np(d["id"]).flatten().tolist()
            rec["eid_0"], rec["eid_1"] = session.name
        else:
            rec["id"] = d["id"]
            rec["eid"] = session.name
        try:
            file_path = urllib.parse.urlsplit(
                d["data_url"], allow_fragments=False
            ).path.strip("/")
            file_path = alfio.remove_uuid_file(file_path, dry=True).as_posix()
            rec["session_path"] = get_session_path(file_path).as_posix()
            rec["rel_path"] = file_path[len(rec["session_path"]) :].strip("/")
        except TypeError:  # no data url is present
            file_path = ""
            rec["session_path"] = file_path
            rec["rel_path"] = file_path
        rec["dataset_type"] = d["dataset_type"]
        rec["revision"] = d["revision"]
        rec["version"] = d["version"]
        rec["collection"] = d["collection"] or ""
        rec["default_revision"] = d["default_revision"] == "True"
        return rec

    if not ses.get("data_dataset_session_related"):
        return session, None
    records = map(_to_record, ses["data_dataset_session_related"])
    index = ["eid_0", "eid_1", "id_0", "id_1"] if int_id else ["eid", "id"]
    datasets = pd.DataFrame(records).set_index(index).sort_index()
    return session, datasets


def datasets2records(datasets, int_id=False) -> pd.DataFrame:
    """Extract datasets DataFrame from one or more Alyx dataset records

    Parameters
    ----------
    datasets : dict, list
        One or more records from the Alyx 'datasets' endpoint

    Returns
    -------
    pd.DataFrame
        Datasets frame
    int_id : bool
        If True, the UUIDs are converted to two int64s

    Examples
    --------
    >>> datasets = ONE().alyx.rest('datasets', 'list', subject='foobar')
    >>> df = datasets2records(datasets)
    """
    records = []

    for d in ensure_list(datasets):
        file_record = next((x for x in d["file_records"] if x["exists"]), None)
        print(file_record)
        if not file_record:
            continue  # Ignore files that are not accessible
        rec = dict(file_size=d["file_size"], hash=d["hash"], exists=True)
        if int_id:
            rec["id_0"], rec["id_1"] = parquet.str2np(d["url"][-36:]).flatten().tolist()
            rec["eid_0"], rec["eid_1"] = (
                parquet.str2np(d["session"][-36:]).flatten().tolist()
            )
        else:
            rec["id"] = d["url"][-36:]
            rec["eid"] = d["session"][-36:]
        data_url = urllib.parse.urlsplit(file_record["data_url"], allow_fragments=False)
        file_path = get_alf_path(data_url.path.strip("/"))
        file_path = alfio.remove_uuid_file(file_path, dry=True).as_posix()
        rec["session_path"] = get_session_path(file_path).as_posix()
        rec["rel_path"] = file_path[len(rec["session_path"]) :].strip("/")
        rec["default_revision"] = d["default_dataset"]
        rec["dataset_type"] = d["dataset_type"]
        rec["version"] = d["version"]
        rec["collection"] = d["collection"] or ""
        rec["revision"] = d["revision"]
        records.append(rec)

    index = ["eid_0", "eid_1", "id_0", "id_1"] if int_id else ["eid", "id"]
    if not records:
        keys = (
            *index,
            "file_size",
            "hash",
            "session_path",
            "rel_path",
            "default_revision",
        )
        return pd.DataFrame(columns=keys).set_index(index)
    return pd.DataFrame(records).set_index(index).sort_index()


def parse_id(method):
    """
    Ensures the input experiment identifier is an experiment UUID string

    Parameters
    ----------
    method : function
        An ONE method whose second arg is an experiment ID

    Returns
    -------
    function
        A wrapper function that parses the ID to the expected string

    Raises
    ------
    ValueError
        Unable to convert input to a valid experiment ID
    """

    @wraps(method)
    def wrapper(self, id, *args, **kwargs):
        eid = self.to_eid(id)
        if eid is None:
            raise ValueError(f'Cannot parse session ID "{id}" (session may not exist)')
        return method(self, eid, *args, **kwargs)

    return wrapper


def refresh(method):
    """
    Refresh cache depending of query_type kwarg
    """

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        mode = kwargs.get("query_type", None)
        if not mode or mode == "auto":
            mode = self.mode
        self.refresh_cache(mode=mode)
        return method(self, *args, **kwargs)

    return wrapper


def validate_date_range(date_range) -> Tuple[pd.Timestamp, pd.Timestamp]:
    """
    Validates and arrange date range in a 2 elements list

    Parameters
    ----------
    date_range : str, datetime.date, datetime.datetime, pd.Timestamp, np.datetime64, list, None
        A single date or tuple/list of two dates.  None represents no bound.

    Returns
    -------
    tuple of pd.Timestamp
        The start and end timestamps

    Examples
    --------
    >>> validate_date_range('2020-01-01')  # On this day
    >>> validate_date_range(datetime.date(2020, 1, 1))
    >>> validate_date_range(np.array(['2022-01-30', '2022-01-30'], dtype='datetime64[D]'))
    >>> validate_date_range(pd.Timestamp(2020, 1, 1))
    >>> validate_date_range(np.datetime64(2021, 3, 11))
    >>> validate_date_range(['2020-01-01'])  # from date
    >>> validate_date_range(['2020-01-01', None])  # from date
    >>> validate_date_range([None, '2020-01-01'])  # up to date

    Raises
    ------
    ValueError
        Size of date range tuple must be 1 or 2
    """
    if date_range is None:
        return

    # Ensure we have exactly two values
    if isinstance(date_range, str) or not isinstance(date_range, Iterable):
        # date_range = (date_range, pd.Timestamp(date_range) + pd.Timedelta(days=1))
        dt = pd.Timedelta(days=1) - pd.Timedelta(milliseconds=1)
        date_range = (date_range, pd.Timestamp(date_range) + dt)
    elif len(date_range) == 1:
        date_range = [date_range[0], pd.Timestamp.max]
    elif len(date_range) != 2:
        raise ValueError

    # For comparisons, ensure both values are pd.Timestamp (datetime, date and datetime64
    # objects will be converted)
    start, end = date_range
    start = start or pd.Timestamp.min  # Convert None to lowest possible date
    end = end or pd.Timestamp.max  # Convert None to highest possible date

    # Convert to timestamp
    if not isinstance(start, pd.Timestamp):
        start = pd.Timestamp(start)
    if not isinstance(end, pd.Timestamp):
        end = pd.Timestamp(end)

    return start, end


def _collection_spec(collection=None, revision=None) -> str:
    """
    Return a template string for a collection/revision regular expression.  Because both are
    optional in the ALF spec, None will match any (including absent), while an empty string will
    match absent.

    Parameters
    ----------
    collection : None, str
        An optional collection regular expression
    revision : None, str
        An optional revision regular expression

    Returns
    -------
    str
        A string format for matching the collection/revision
    """
    spec = ""
    for value, default in zip(
        (collection, revision), ("{collection}/", "#{revision}#/")
    ):
        if not value:
            default = f"({default})?" if value is None else ""
        spec += default
    return spec


def _file_spec(**kwargs):
    """
    Return a template string for a ALF dataset regular expression.  Because 'namespace',
    'timescale', and 'extra' are optional None will match any (including absent).  This function
    removes the regex flags from the file spec string that make certain parts optional.

    TODO an empty string should only match absent; this could be achieved by removing parts from
     spec string

    Parameters
    ----------
    namespace : None, str
        If namespace is not None, the namespace section of the returned file spec will not be
        optional.
    timescale : None, str
        If timescale is not None, the namespace section of the returned file spec will not be
        optional.
    extra : None, str
        If extra is not None, the namespace section of the returned file spec will not be
        optional.

    Returns
    -------
    str
        A string format for matching an ALF dataset
    """
    OPTIONAL = {"namespace": "?", "timescale": "?", "extra": "*"}
    filespec = FILE_SPEC
    for k, v in kwargs.items():
        if k in OPTIONAL and v is not None:
            i = filespec.find(k) + len(k)
            i += filespec[i:].find(OPTIONAL[k])
            filespec = filespec[:i] + filespec[i:].replace(OPTIONAL[k], "", 1)
    return filespec


def filter_datasets(
    all_datasets,
    filename=None,
    collection=None,
    revision=None,
    revision_last_before=True,
    assert_unique=True,
    wildcards=False,
):
    """
    Filter the datasets cache table by the relative path (dataset name, collection and revision).
    When None is passed, all values will match.  To match on empty parts, use an empty string.
    When revision_last_before is true, None means return latest revision.

    Parameters
    ----------
    all_datasets : pandas.DataFrame
        A datasets cache table
    filename : str, dict, None
        A filename str or a dict of alf parts.  Regular expressions permitted.
    collection : str, None
        A collection string.  Regular expressions permitted.
    revision : str, None
        A revision string to match.  If revision_last_before is true, regular expressions are
        not permitted.
    revision_last_before : bool
        When true and no exact match exists, the (lexicographically) previous revision is used
        instead.  When false the revision string is matched like collection and filename,
        with regular expressions permitted.
    assert_unique : bool
        When true an error is raised if multiple collections or datasets are found
    wildcards : bool
        If true, use unix shell style matching instead of regular expressions

    Returns
    -------
    pd.DataFrame
        A slice of all_datasets that match the filters

    Examples
    --------
    Filter by dataset name and collection

    >>> datasets = filter_datasets(all_datasets, '.*spikes.times.*', 'alf/probe00')

    Filter datasets not in a collection

    >>> datasets = filter_datasets(all_datasets, collection='')

    Filter by matching revision

    >>> datasets = filter_datasets(all_datasets, 'spikes.times.npy',
    ...                            revision='2020-01-12', revision_last_before=False)

    Filter by filename parts

    >>> datasets = filter_datasets(all_datasets, dict(object='spikes', attribute='times'))
    """

    def text_is_in(cell, text=None):
        if text is None:
            return True
        if cell is None:
            return False
        if re.match(text, cell):
            return True
        return False

    collection_filter = all_datasets["collection"].apply(text_is_in, text=collection)
    filename_filter = all_datasets["dataset_type"].apply(text_is_in, text=filename)

    match = all_datasets[collection_filter & filename_filter]
    return filter_revision_last_before(match, revision, assert_unique=assert_unique)

    # Create a regular expression string to match relative path against
    filename = filename or {}
    regex_args = {"collection": collection}
    spec_str = _collection_spec(collection, None if revision_last_before else revision)

    if isinstance(filename, dict):
        spec_str += _file_spec(**filename)
        regex_args.update(**filename)
    else:
        # Convert to regex is necessary and assert end of string
        filename = [
            fnmatch.translate(x) if wildcards else x + "$"
            for x in ensure_list(filename)
        ]
        spec_str += "|".join(filename)

    # If matching revision name, add to regex string
    if not revision_last_before:
        regex_args.update(revision=revision)

    for k, v in regex_args.items():
        if v is None:
            continue
        if wildcards:
            # Convert to regex, remove \\Z which asserts end of string
            v = (fnmatch.translate(x).replace("\\Z", "") for x in ensure_list(v))
        if not isinstance(v, str):
            regex_args[k] = "|".join(v)  # logical OR

    # Build regex string

    pattern = alf_regex("^" + spec_str, **regex_args)
    match = all_datasets[all_datasets["rel_path"].str.match(pattern)]
    if len(match) == 0 or not (revision_last_before or assert_unique):
        return match

    revisions = [rel_path_parts(x)[1] or "" for x in match.rel_path.values]
    if assert_unique:
        collections = set(rel_path_parts(x)[0] or "" for x in match.rel_path.values)
        if len(collections) > 1:
            _list = '"' + '", "'.join(collections) + '"'
            raise alferr.ALFMultipleCollectionsFound(_list)
        if not revision_last_before:
            if filename and len(match) > 1:
                _list = '"' + '", "'.join(match["rel_path"]) + '"'
                raise alferr.ALFMultipleObjectsFound(_list)
            if len(set(revisions)) > 1:
                _list = '"' + '", "'.join(set(revisions)) + '"'
                raise alferr.ALFMultipleRevisionsFound(_list)
            else:
                return match
        elif filename and len(set(revisions)) != len(revisions):
            _list = '"' + '", "'.join(match["rel_path"]) + '"'
            raise alferr.ALFMultipleObjectsFound(_list)

    return filter_revision_last_before(match, revision, assert_unique=assert_unique)


def filter_revision_last_before(datasets, revision=None, assert_unique=True):
    """
    Filter datasets by revision, returning previous revision in ordered list if revision
    doesn't exactly match.

    Parameters
    ----------
    datasets : pandas.DataFrame
        A datasets cache table
    revision : str
        A revision string to match (regular expressions not permitted)
    assert_unique : bool
        When true an alferr.ALFMultipleRevisionsFound exception is raised when multiple
        default revisions are found; an alferr.ALFError when no default revision is found

    Returns
    -------
    pd.DataFrame
        A datasets DataFrame with 0 or 1 row per unique dataset
    """
    if revision is None:
        return datasets[datasets["default_revision"]]
    return datasets[datasets["revision"] == revision]

    def _last_before(df):
        """Takes a DataFrame with only one dataset and multiple revisions, returns matching row"""
        if revision is None and "default_revision" in df.columns:
            if assert_unique and sum(df.default_revision) > 1:
                revisions = df["revision"][df.default_revision.values]
                rev_list = '"' + '", "'.join(revisions) + '"'
                raise alferr.ALFMultipleRevisionsFound(rev_list)
            if sum(df.default_revision) == 1:
                return df[df.default_revision]
            # default_revision column all False; default doesn't isn't copied to remote repository
            dset_name = df["rel_path"].iloc[0]
            if assert_unique:
                raise alferr.ALFError(f"No default revision for dataset {dset_name}")
            else:
                logger.warning(
                    f"No default revision for dataset {dset_name}; using most recent"
                )
        # Compare revisions lexicographically
        if assert_unique and len(df["revision"].unique()) > 1:
            rev_list = '"' + '", "'.join(df["revision"].unique()) + '"'
            raise alferr.ALFMultipleRevisionsFound(rev_list)
        # Square brackets forces 1 row DataFrame returned instead of Series
        idx = index_last_before(df["revision"].tolist(), revision)
        # return df.iloc[slice(0, 0) if idx is None else [idx], :]
        return df.iloc[slice(0, 0) if idx is None else [idx], :]

    with pd.option_context("mode.chained_assignment", None):  # FIXME Explicitly copy?
        datasets["revision"] = [rel_path_parts(x)[1] or "" for x in datasets.rel_path]
    groups = datasets.rel_path.str.replace("#.*#/", "", regex=True).values
    grouped = datasets.groupby(groups, group_keys=False)
    return grouped.apply(_last_before)


def index_last_before(revisions: List[str], revision: Optional[str]) -> Optional[int]:
    """
    Returns the index of string that occurs directly before the provided revision string when
    lexicographic sorted.  If revision is None, the index of the most recent revision is returned.

    Parameters
    ----------
    revisions : list of strings
        A list of revision strings
    revision : None, str
        The revision string to match on

    Returns
    -------
    int, None
        Index of revision before matching string in sorted list or None

    Examples
    --------
    >>> idx = index_last_before([], '2020-08-01')
    """
    if len(revisions) == 0:
        return  # No revisions, just return
    revisions_sorted = sorted(revisions, reverse=True)
    if revision is None:  # Return most recent revision
        return revisions.index(revisions_sorted[0])
    lt = np.array(revisions_sorted) <= revision
    return revisions.index(revisions_sorted[lt.argmax()]) if any(lt) else None


def autocomplete(term, search_terms) -> str:
    """
    Validate search term and return complete name, e.g. autocomplete('subj') == 'subject'
    """
    term = term.lower()
    # Check if term already complete
    if term in search_terms:
        return term
    full_key = (x for x in search_terms if x.lower().startswith(term))
    key_ = next(full_key, None)
    if not key_:
        raise ValueError(f'Invalid search term "{term}", see `one.search_terms()`')
    elif next(full_key, None):
        raise ValueError(f'Ambiguous search term "{term}"')
    return key_


def ensure_list(value):
    """Ensure input is a list"""
    return (
        [value]
        if isinstance(value, (str, dict)) or not isinstance(value, Iterable)
        else value
    )


class LazyId(Mapping):
    """
    Using a paginated response object or list of session records, extracts eid string when required
    """

    def __init__(self, pg):
        self._pg = pg

    def __getitem__(self, item):
        return self.ses2eid(self._pg.__getitem__(item))

    def __len__(self):
        return self._pg.__len__()

    def __iter__(self):
        return map(self.ses2eid, self._pg.__iter__())

    @staticmethod
    def ses2eid(ses):
        """

        Parameters
        ----------
        ses : one.webclient._PaginatedResponse, dict, list
            A collection of Alyx REST sessions endpoint records

        Returns
        -------
        str, list
            One or more experiment ID strings
        """
        if isinstance(ses, list):
            return [LazyId.ses2eid(x) for x in ses]
        else:
            return ses.get("id", None) or ses["url"].split("/").pop()


def cache_int2str(table: pd.DataFrame) -> pd.DataFrame:
    """Convert int ids to str ids for cache table.

    Parameters
    ----------
    table : pd.DataFrame
        A cache table (from One._cache)

    """
    # Convert integer uuids to str uuids
    if table.index.nlevels < 2 or not any(x.endswith("_0") for x in table.index.names):
        return table
    table = table.reset_index()
    int_cols = table.filter(regex=r"_\d{1}$").columns.sort_values()
    assert not len(int_cols) % 2, "expected even number of columns ending in _0 or _1"
    names = sorted(set(c.rsplit("_", 1)[0] for c in int_cols.values))
    for i, name in zip(range(0, len(int_cols), 2), names):
        table[name] = parquet.np2str(table[int_cols[i : i + 2]])
    table = table.drop(int_cols, axis=1).set_index(names)
    return table


def patch_cache(table: pd.DataFrame, min_api_version=None) -> pd.DataFrame:
    """Reformat older cache tables to comply with this version of ONE.

    Currently this function will 1. convert integer UUIDs to string UUIDs; 2. rename the 'project'
    column to 'projects'.

    Parameters
    ----------
    table : pd.DataFrame
        A cache table (from One._cache).
    min_api_version : str
        The minimum API version supported by this cache table.
    """
    min_version = version.parse(min_api_version or "0.0.0")
    table = cache_int2str(table)
    # Rename project column
    if min_version < version.Version("1.13.0") and "project" in table.columns:
        table.rename(columns={"project": "projects"}, inplace=True)
    return table



"""I/O functions for ALyx Files.

Provides support for time-series reading and interpolation as per the specifications
For a full overview of the scope of the format, see:

https://int-brain-lab.github.io/ONE/alf_intro.html
"""

import json
import copy
import logging
import os
import re
from fnmatch import fnmatch
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
import yaml

from iblutil.util import Bunch
from iblutil.io import parquet
from iblutil.io import jsonable
from .exceptions import ALFObjectNotFound
from . import files, spec
from .spec import FILE_SPEC

_logger = logging.getLogger(__name__)


class AlfBunch(Bunch):
    """A dict-like object that supports dot indexing and conversion to DataFrame"""

    @property
    def check_dimensions(self):
        """int: 0 for consistent dimensions, 1 for inconsistent dimensions"""
        return check_dimensions(self)

    def append(self, b, inplace=False):
        """
        Appends one bunch to another, key by key

        Parameters
        ----------
        b : Bunch, dict
            A Bunch of data to append
        inplace : bool
            If true, the data are appended in place, otherwise a copy is returned

        Returns
        -------
        ALFBunch, None
            An ALFBunch with the data appended, or None if inplace is True
        """
        # default is to return a copy
        if inplace:
            a = self
        else:
            a = AlfBunch(copy.deepcopy(self))
        # handles empty bunches for convenience if looping
        if b == {}:
            return a
        if a == {}:
            return AlfBunch(b)
        # right now supports only strictly matching keys. Will implement other cases as needed
        if set(a.keys()) != set(b.keys()):
            raise NotImplementedError('Append bunches only works with strictly matching keys'
                                      'For more complex merges, convert to pandas dataframe.')
        # do the merge; only concatenate lists and np arrays right now
        for k in a:
            if isinstance(a[k], np.ndarray):
                a[k] = np.concatenate((a[k], b[k]), axis=0)
            elif isinstance(a[k], list):
                a[k].extend(b[k])
            else:
                _logger.warning(f'bunch key "{k}" is a {a[k].__class__}. I don\'t know how to'
                                f' handle that. Use pandas for advanced features')
        if a.check_dimensions != 0:
            print_sizes = '\n'.join(f'{v.shape},\t{k}' for k, v in a.items())
            _logger.warning(f'Inconsistent dimensions for object: \n{print_sizes}')

        return a

    def to_df(self) -> pd.DataFrame:
        """Return DataFrame with data keys as columns"""
        return dataframe(self)

    @staticmethod
    def from_df(df) -> 'AlfBunch':
        data = dict(zip(df.columns, df.values.T))
        split_keys = sorted(x for x in data.keys() if re.match(r'.+?_[01]$', x))
        for x1, x2 in zip(*[iter(split_keys)] * 2):
            data[x1[:-2]] = np.c_[data.pop(x1), data.pop(x2)]
        return AlfBunch(data)


def dataframe(adict):
    """
    Converts an Bunch conforming to size conventions into a pandas DataFrame.
    For 2-D arrays, stops at 10 columns per attribute.

    Parameters
    ----------
    adict : dict, Bunch
        A dict-like object of data to convert to DataFrame

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame of data
    """
    if check_dimensions(adict) != 0:
        raise ValueError('Can only convert to DataFrame objects with consistent size')
    # easy case where there are only vectors
    if all([len(adict[k].shape) == 1 for k in adict]):
        return pd.DataFrame(adict)
    # pandas has trouble with 2d data, chop it off with a limit of 10 columns per dataset
    df = pd.DataFrame()
    for k in adict.keys():
        if adict[k].ndim == 1:
            df[k] = adict[k]
        elif adict[k].ndim == 2 and adict[k].shape[1] == 1:
            df[k] = adict[k][:, 0]
        elif adict[k].ndim == 2:
            for i in np.arange(adict[k].shape[1]):
                df[f"{k}_{i}"] = adict[k][:, i]
                if i == 9:
                    break
        else:
            _logger.warning(f'{k} attribute is 3D or more and won\'t convert to dataframe')
            continue
    return df


def _find_metadata(file_alf) -> Path:
    """
    File path for an existing meta-data file for an alf_file

    Parameters
    ----------
    file_alf : str, pathlib.Path
        A path of existing ALF

    Returns
    -------
    pathlib.Path
        Path of meta-data file if exists
    """
    file_alf = Path(file_alf)
    ns, obj = file_alf.name.split('.')[:2]
    return next(file_alf.parent.glob(f'{ns}.{obj}*.metadata*.json'), None)


def read_ts(filename):
    """
    Load time-series from ALF format

    Parameters
    ----------
    filename : str, pathlib.Path
        An ALF path whose values to load

    Returns
    -------
    numpy.ndarray
        An array of timestamps belonging to the ALF path object
    numpy.ndarray
         An array of values in filename

    Examples
    --------
    >>> t, d = read_ts(filename)
    """
    if not isinstance(filename, Path):
        filename = Path(filename)

    # alf format is object.attribute.extension, for example '_ibl_wheel.position.npy'
    _, obj, attr, *_, ext = files.filename_parts(filename.parts[-1])

    try:
        # looking for matching object with attribute timestamps: '_ibl_wheel.timestamps.npy'
        (time_file,), _ = filter_by(filename.parent, object=obj, attribute='times*', extension=ext)
        assert time_file
    except (ValueError, AssertionError):
        name = spec.to_alf(obj, attr, ext)
        raise FileNotFoundError(name + ' not found! No time-scale for ' + str(filename))

    ts = np.load(filename.parent / time_file)
    val = np.load(filename)
    # Ensure timestamps
    return ts2vec(ts, val.shape[0]), _ensure_flat(val)


def _ensure_flat(arr):
    """
    Given a single column array, returns a flat vector.  Other shapes are returned unchanged.

    Parameters
    ----------
    arr : numpy.array
        An array with shape (n, 1)

    Returns
    -------
    numpy.ndarray
        A vector with shape (n,)
    """
    return arr.flatten() if arr.ndim == 2 and arr.shape[1] == 1 else arr


def ts2vec(ts: np.ndarray, n_samples: int) -> np.ndarray:
    """
    Interpolate a continuous timeseries of the shape (2, 2)

    Parameters
    ----------
    ts : numpy.array
        a 2x2 numpy array of the form (sample, ts)
    n_samples : int
        Number of samples; i.e. the size of the resulting vector

    Returns
    -------
    numpy.ndarray
        A vector of interpolated timestamps
    """
    if len(ts.shape) == 1:
        return ts
    elif ts.ndim == 2 and ts.shape[1] == 1:
        return ts.flatten()  # Deal with MATLAB single column array
    if ts.ndim > 2 or ts.shape[1] != 2:
        raise ValueError('Array shape should be (2, 2)')
    # Linearly interpolate the times
    x = np.arange(n_samples)
    return np.interp(x, ts[:, 0], ts[:, 1])


def check_dimensions(dico):
    """
    Test for consistency of dimensions as per ALF specs in a dictionary.

    Alf broadcasting rules: only accepts consistent dimensions for a given axis
    a dimension is consistent with another if it's empty, 1, or equal to the other arrays
    dims [a, 1],  [1, b] and [a, b] are all consistent, [c, 1] is not

    Parameters
    ----------
    dico : ALFBunch, dict
        Dictionary containing data

    Returns
    -------
    int
        Status 0 for consistent dimensions, 1 for inconsistent dimensions
    """
    supported = (np.ndarray, pd.DataFrame)  # Data types that have a shape attribute
    shapes = [dico[lab].shape for lab in dico
              if isinstance(dico[lab], supported) and not lab.startswith('timestamps')]
    first_shapes = [sh[0] for sh in shapes]
    # Continuous timeseries are permitted to be a (2, 2)
    timeseries = [k for k, v in dico.items()
                  if k.startswith('timestamps') and isinstance(v, np.ndarray)]
    if any(timeseries):
        for key in timeseries:
            if dico[key].ndim == 1 or (dico[key].ndim == 2 and dico[key].shape[1] == 1):
                # Should be vector with same length as other attributes
                first_shapes.append(dico[key].shape[0])
            elif dico[key].ndim > 1 and dico[key].shape != (2, 2):
                return 1  # ts not a (2, 2) arr or a vector

    ok = len(first_shapes) == 0 or set(first_shapes).issubset({max(first_shapes), 1})
    return int(ok is False)


def load_file_content(fil):
    """
    Returns content of files. Designed for very generic file formats:
    so far supported contents are `json`, `npy`, `csv`, `(h)tsv`, `ssv`, `jsonable`

    Parameters
    ----------
    fil : str, pathlib.Path
        File to read

    Returns
    -------
    any
        Array/json/pandas dataframe depending on format
    """
    if not fil:
        return
    fil = Path(fil)
    if fil.stat().st_size == 0:
        return
    if fil.suffix == '.csv':
        return pd.read_csv(fil).squeeze('columns')
    if fil.suffix == '.json':
        try:
            with open(fil) as _fil:
                return json.loads(_fil.read())
        except Exception as e:
            _logger.error(e)
            return None
    if fil.suffix == '.jsonable':
        return jsonable.read(fil)
    if fil.suffix == '.npy':
        return _ensure_flat(np.load(file=fil, allow_pickle=True))
    if fil.suffix == '.pqt':
        return parquet.load(fil)[0]
    if fil.suffix == '.ssv':
        return pd.read_csv(fil, delimiter=' ').squeeze('columns')
    if fil.suffix in ('.tsv', '.htsv'):
        return pd.read_csv(fil, delimiter='\t').squeeze('columns')
    if fil.suffix in ('.yml', '.yaml'):
        with open(fil, 'r') as _fil:
            return yaml.safe_load(_fil)
    return Path(fil)


def _ls(alfpath, object=None, **kwargs) -> (list, tuple):
    """
    Given a path, an object and a filter, returns all files and associated attributes

    Parameters
    ----------
    alfpath : str, pathlib.Path
        The folder to list
    object : str, list
        An ALF object name to filter by
    wildcards : bool
        If true uses unix shell style pattern matching, otherwise uses regular expressions
    kwargs : dict
        Other ALF parts to filter, including namespace, attribute, etc.

    Returns
    -------
    list
        A list of ALF paths
    tuple
        A tuple of ALF attributes corresponding to the file paths

    Raises
    ------
    ALFObjectNotFound
        No matching ALF object was found in the alfpath directory
    """
    alfpath = Path(alfpath)
    if not alfpath.exists():
        files_alf = attributes = None
    elif alfpath.is_dir():
        if object is None:
            # List all ALF files
            files_alf, attributes = filter_by(alfpath, **kwargs)
        else:
            files_alf, attributes = filter_by(alfpath, object=object, **kwargs)
    else:
        object = files.filename_parts(alfpath.name)[1]
        alfpath = alfpath.parent
        files_alf, attributes = filter_by(alfpath, object=object, **kwargs)

    # raise error if no files found
    if not files_alf:
        err_str = f'object "{object}" ' if object else 'ALF files'
        raise ALFObjectNotFound(f'No {err_str} found in {alfpath}')

    return [alfpath.joinpath(f) for f in files_alf], attributes


def iter_sessions(root_dir):
    """
    Recursively iterate over session paths in a given directory.

    Parameters
    ----------
    root_dir : str, pathlib.Path
        The folder to look for sessions.

    Yields
    -------
    pathlib.Path
        The next session path in lexicographical order.
    """
    if spec.is_session_path(root_dir):
        yield root_dir
    for path in sorted(Path(root_dir).rglob('*')):
        if path.is_dir() and spec.is_session_path(path):
            yield path


def iter_datasets(session_path):
    """
    Iterate over all files in a session, and yield relative dataset paths.

    Parameters
    ----------
    session_path : str, pathlib.Path
        The folder to look for datasets.

    Yields
    -------
    pathlib.Path
        The next dataset path (relative to the session path) in lexicographical order.
    """
    for p in sorted(Path(session_path).rglob('*.*')):
        if not p.is_dir() and spec.is_valid(p.name):
            yield p.relative_to(session_path)


def exists(alfpath, object, attributes=None, **kwargs) -> bool:
    """
    Test if ALF object and optionally specific attributes exist in the given path

    Parameters
    ----------
    alfpath : str, pathlib.Path
        The folder to look into
    object : str
        ALF object name
    attributes : str, list
        Wanted attributes
    wildcards : bool
        If true uses unix shell style pattern matching, otherwise uses regular expressions
    kwargs : dict
        Other ALF parts to filter by

    Returns
    -------
    bool
        For multiple attributes, returns True only if all attributes are found
    """

    # if the object is not found, return False
    try:
        _, attributes_found = _ls(alfpath, object, **kwargs)
    except (FileNotFoundError, ALFObjectNotFound):
        return False

    # if object found and no attribute provided, True
    if not attributes:
        return True

    # if attributes provided, test if all are found
    if isinstance(attributes, str):
        attributes = [attributes]
    attributes_found = set(part[2] for part in attributes_found)
    return set(attributes).issubset(attributes_found)


def load_object(alfpath, object=None, short_keys=False, **kwargs):
    """Reads all files sharing the same object name.

    For example, if the file provided to the function is `spikes.times`, the function will
    load `spikes.times`, `spikes.clusters`, `spikes.depths`, `spike.amps` in a dictionary
    whose keys will be `times`, `clusters`, `depths`, `amps`

    Full Reference here: https://int-brain-lab.github.io/ONE/alf_intro.html

    Simplified example: _namespace_object.attribute_timescale.part1.part2.extension

    Parameters
    ----------
    alfpath : str, pathlib.Path, list
        Any ALF path pertaining to the object OR directory containing ALFs OR list of paths
    object : str, list, None
        The ALF object(s) to filter by.  If a directory is provided and object is None, all valid
        ALF files returned
    short_keys : bool
        By default, the output dictionary keys will be compounds of attributes, timescale and
        any eventual parts separated by a dot. Use True to shorten the keys to the attribute
        and timescale
    wildcards : bool
        If true uses unix shell style pattern matching, otherwise uses regular expressions
    kwargs : dict
        Other ALF parts to filter by

    Returns
    -------
    AlfBunch
        An ALFBunch (dict-like) of all attributes pertaining to the object

    Examples
    --------
        Load 'spikes' object

        >>> spikes = load_object('full/path/to/my/alffolder/', 'spikes')

        Load 'trials' object under the 'ibl' namespace

        >>> trials = load_object('/subject/2021-01-01/001', 'trials', namespace='ibl')
    """
    if isinstance(alfpath, (Path, str)):
        if Path(alfpath).is_dir() and object is None:
            raise ValueError('If a directory is provided, the object name should be provided too')
        files_alf, parts = _ls(alfpath, object, **kwargs)
    else:  # A list of paths allows us to load an object from different revisions
        files_alf = alfpath
        parts = [files.filename_parts(x.name) for x in files_alf]
        assert len(set(p[1] for p in parts)) == 1
        object = next(x[1] for x in parts)
    # Take attribute and timescale from parts list
    attributes = [p[2] if not p[3] else '_'.join(p[2:4]) for p in parts]
    if not short_keys:  # Include extra parts in the keys
        attributes = [attr + ('.' + p[4] if p[4] else '') for attr, p in zip(attributes, parts)]
        # TODO List duplicates; raise ALFError
    assert len(set(attributes)) == len(attributes), (
        f'multiple object {object} with the same attribute in {alfpath}, restrict parts/namespace')
    out = AlfBunch({})

    # load content for each file
    for fil, att in zip(files_alf, attributes):
        # if there is a corresponding metadata file, read it:
        meta_data_file = _find_metadata(fil)
        # if this is the actual meta-data file, skip and it will be read later
        if meta_data_file == fil:
            continue
        out[att] = load_file_content(fil)
        if meta_data_file:
            meta = load_file_content(meta_data_file)
            # the columns keyword splits array along the last dimension
            if 'columns' in meta.keys():
                out.update({v: out[att][::, k] for k, v in enumerate(meta['columns'])})
                out.pop(att)
                meta.pop('columns')
            # if there is other stuff in the dictionary, save it, otherwise disregard
            if meta:
                out[att + 'metadata'] = meta
    if 'table' in out.keys():  # Merge 'table' dataframe into bunch
        out.update(AlfBunch.from_df(out.pop('table')))
    status = out.check_dimensions
    timeseries = [k for k in out.keys() if 'timestamps' in k]
    if any(timeseries) and len(out.keys()) > len(timeseries) and status == 0:
        # Get length of one of the other arrays
        ignore = ('timestamps', 'meta')
        n_samples = next(v for k, v in out.items() if not any(x in k for x in ignore)).shape[0]
        for key in timeseries:
            # Expand timeseries if necessary
            out[key] = ts2vec(out[key], n_samples)
    if status != 0:
        print_sizes = '\n'.join(f'{v.shape},\t{k}' for k, v in out.items())
        _logger.warning(f'Inconsistent dimensions for object: {object} \n{print_sizes}')
    return out


def save_object_npy(alfpath, dico, object, parts=None, namespace=None, timescale=None) -> list:
    """
    Saves a dictionary in `ALF format`_ using object as object name and dictionary keys as
    attribute names. Dimensions have to be consistent.

    Simplified ALF example: _namespace_object.attribute.part1.part2.extension

    Parameters
    ----------
    alfpath : str, pathlib.Path
        Path of the folder to save data to
    dico : dict
        Dictionary to save to npy; keys correspond to ALF attributes
    object : str
        Name of the object to save
    parts : str, list, None
        Extra parts to the ALF name
    namespace : str, None
        The optional namespace of the object
    timescale : str, None
        The optional timescale of the object

    Returns
    -------
    list
        List of written files

    Examples
    --------
    >>> spikes = {'times': np.arange(50), 'depths': np.random.random(50)}
    >>> files = save_object_npy('/path/to/my/alffolder/', spikes, 'spikes')

    .. _ALF format:
        https://int-brain-lab.github.io/ONE/alf_intro.html
    """
    alfpath = Path(alfpath)
    status = check_dimensions(dico)
    if status != 0:
        raise ValueError('Dimensions are not consistent to save all arrays in ALF format: ' +
                         str([(k, v.shape) for k, v in dico.items()]))
    out_files = []
    for k, v in dico.items():
        out_file = alfpath / spec.to_alf(object, k, 'npy',
                                         extra=parts, namespace=namespace, timescale=timescale)
        np.save(out_file, v)
        out_files.append(out_file)
    return out_files


def save_metadata(file_alf, dico) -> None:
    """Writes a meta data file matching a current ALF file object.

    For example given an alf file `clusters.ccfLocation.ssv` this will write a dictionary in JSON
    format in `clusters.ccfLocation.metadata.json`

    Reserved keywords:
        - **columns**: column names for binary tables.
        - **row**: row names for binary tables.
        - **unit**

    Parameters
    ----------
    file_alf : str, pathlib.Path
        Full path to the alf object
    dico : dict, ALFBunch
        Dictionary containing meta-data
    """
    assert spec.is_valid(file_alf.parts[-1]), 'ALF filename not valid'
    file_meta_data = file_alf.parent / (file_alf.stem + '.metadata.json')
    with open(file_meta_data, 'w+') as fid:
        fid.write(json.dumps(dico, indent=1))


def remove_uuid_file(file_path, dry=False) -> Path:
    """
    Renames a file without the UUID and returns the new pathlib.Path object
    """
    if isinstance(file_path, str):
        file_path = Path(file_path)
    name_parts = file_path.name.split('.')
    if not spec.is_uuid_string(name_parts[-2]):
        return file_path
    name_parts.pop(-2)
    new_path = file_path.parent.joinpath('.'.join(name_parts))
    if not dry and file_path.exists():
        file_path.replace(new_path)
    return new_path


def remove_uuid_recursive(folder, dry=False) -> None:
    """
    Within a folder, recursive renaming of all files to remove UUID
    """
    for fn in Path(folder).rglob('*.*'):
        print(remove_uuid_file(fn, dry=dry))


def next_num_folder(session_date_folder: Union[str, Path]) -> str:
    """Return the next number for a session given a session_date_folder"""
    session_date_folder = Path(session_date_folder)
    if not session_date_folder.exists():
        return '001'
    session_nums = [
        int(x.name) for x in session_date_folder.iterdir()
        if x.is_dir() and not x.name.startswith('.') and x.name.isdigit()
    ]
    out = f'{max(session_nums or [0]) + 1:03d}'
    assert len(out) == 3, 'ALF spec does not support session numbers > 999'
    return out


def remove_empty_folders(folder: Union[str, Path]) -> None:
    """Will iteratively remove any children empty folders"""
    all_folders = sorted(x for x in Path(folder).rglob('*') if x.is_dir())
    for f in reversed(all_folders):  # Reversed sorted ensures we remove deepest first
        try:
            f.rmdir()
        except Exception:
            continue


def filter_by(alf_path, wildcards=True, **kwargs):
    """
    Given a path and optional filters, returns all ALF files and their associated parts. The
    filters constitute a logical AND.  For all but `extra`, if a list is provided, one or more
    elements must match (a logical OR).

    Parameters
    ----------
    alf_path : str, pathlib.Path
        A path to a folder containing ALF datasets
    wildcards : bool
        If true, kwargs are matched as unix-style patterns, otherwise as regular expressions
    object : str, list
        Filter by a given object (e.g. 'spikes')
    attribute : str, list
        Filter by a given attribute (e.g. 'intervals')
    extension : str, list
        Filter by extension (e.g. 'npy')
    namespace : str, list
        Filter by a given namespace (e.g. 'ibl') or None for files without one
    timescale : str, list
        Filter by a given timescale (e.g. 'bpod') or None for files without one
    extra : str, list
        Filter by extra parameters (e.g. 'raw') or None for files without extra parts
        NB: Wild cards not permitted here.

    Returns
    -------
    alf_files : str
        A Path to a directory containing ALF files
    attributes : list of dicts
        A list of parsed file parts

    Examples
    --------
    Filter files with universal timescale

    >>> filter_by(alf_path, timescale=None)

    Filter files by a given ALF object

    >>> filter_by(alf_path, object='wheel')

    Filter using wildcard, e.g. 'wheel' and 'wheelMoves' ALF objects

    >>> filter_by(alf_path, object='wh*')

    Filter all intervals that are in bpod time

    >>> filter_by(alf_path, attribute='intervals', timescale='bpod')

    Filter all files containing either 'intervals' OR 'timestamps' attributes

    >>> filter_by(alf_path, attribute=['intervals', 'timestamps'])

    Filter all files using a regular expression

    >>> filter_by(alf_path, object='^wheel.*', wildcards=False)
    >>> filter_by(alf_path, object=['^wheel$', '.*Moves'], wildcards=False)
    """
    alf_files = [f for f in os.listdir(alf_path) if spec.is_valid(f)]
    attributes = [files.filename_parts(f, as_dict=True) for f in alf_files]

    if kwargs:
        # Validate keyword arguments against regex group names
        invalid = kwargs.keys() - spec.regex(FILE_SPEC).groupindex.keys()
        if invalid:
            raise TypeError('%s() got an unexpected keyword argument "%s"'
                            % (__name__, set(invalid).pop()))

        # # Ensure 'extra' input is a list; if str split on dot
        if 'extra' in kwargs and isinstance(kwargs['extra'], str):
            kwargs['extra'] = kwargs['extra'].split('.')

        def _match(part, pattern, split=None):
            if pattern is None or part is None:
                # If either is None, both should be None to match
                return pattern is part
            elif split:
                # Check all provided extra fields match those in ALF
                return all(elem in part.split(split) for elem in pattern if elem)
            elif not isinstance(pattern, str):
                if wildcards:
                    return any(_match(part, x, split) for x in pattern)
                else:
                    return re.match('|'.join(pattern), part) is not None
            else:
                # Check given attribute matches, allowing wildcards
                return fnmatch(part, pattern) if wildcards else re.match(pattern, part) is not None

        # Iterate over ALF files
        for file, attr in zip(alf_files.copy(), attributes.copy()):
            for k, v in kwargs.items():  # Iterate over attributes
                match = _match(attr[k], v, '.' if k == 'extra' else None)

                if not match:  # Remove file from list and move on to next file
                    alf_files.remove(file)
                    attributes.remove(attr)
                    break

    return alf_files, [tuple(attr.values()) for attr in attributes]
